#include <fstream>
#include <algorithm>
#include <ReTexture/Texture.h>
#define STB_IMAGE_IMPLEMENTATION
#include <stb_image.h>
#define STB_IMAGE_RESIZE_IMPLEMENTATION
#include <stb_image_resize2.h>

namespace ReTexture {

    SoftwareTexture::SoftwareTexture(const std::string &file, const TextureFlag& flags) {
        load(file);
        if (flags & TextureFlag::INVERT_Y)
            invertY();

        this->flags = flags;
    }

    SoftwareTexture::SoftwareTexture(const std::string &file) {
        *this = SoftwareTexture(file, {});
    }

    SoftwareTexture::SoftwareTexture(const std::vector<unsigned char>& pixel_data, TextureFormat format, unsigned int width, unsigned int height) {
        pixelData = pixel_data;
        this->width = width;
        this->height = height;
        this->format = format;
        this->flags = TextureFlag::NONE;
    }


    void SoftwareTexture::load(const std::string &file) {
        std::ifstream ifStream(file, std::ios::in | std::ios::binary);
        unsigned char bmpFileHeader[14];

        ifStream.read(reinterpret_cast<char *>(bmpFileHeader), 14);
        if (bmpFileHeader[0] == 'B' && bmpFileHeader[1] == 'M') {
            ifStream.close();
            loadBMP(file);
            return;
        }

        //TODO determine if it's a PNG using the file header instead.
        if (file.ends_with(".png"))
            loadPNG(file);
    }

    void SoftwareTexture::loadBMP(const std::string &file) {
        std::ifstream bmpFile(file, std::ios::in | std::ios::binary);
        if (!bmpFile.is_open())
            return;

        unsigned char bmpFileHeader[14];
        unsigned char bmpInfoHeader[40];

        bmpFile.read(reinterpret_cast<char *>(bmpFileHeader), 14);
        bmpFile.read(reinterpret_cast<char *>(bmpInfoHeader), 40);

        width = bmpInfoHeader[4] + (bmpInfoHeader[5] << 8) + (bmpInfoHeader[6] << 16) + (bmpInfoHeader[7] << 24);
        height = bmpInfoHeader[8] + (bmpInfoHeader[9] << 8) + (bmpInfoHeader[10] << 16) + (bmpInfoHeader[11] << 24);

        int rowPadded = (width * 3 + 3) & (~3);
        pixelData.resize(width * height * 3);

        std::vector<unsigned char> rowData(rowPadded);
        for (int y = height - 1; y >= 0; --y) {
            bmpFile.read(reinterpret_cast<char *>(rowData.data()), rowPadded);
            for (int x = 0; x < width; ++x) {
                pixelData[(y * width + x) * 3 + 2] = rowData[x * 3 + 0];
                pixelData[(y * width + x) * 3 + 1] = rowData[x * 3 + 1];
                pixelData[(y * width + x) * 3 + 0] = rowData[x * 3 + 2];
            }
        }
        bmpFile.close();
        format = TextureFormat::RGB;
    }

    void SoftwareTexture::loadPNG(const std::string &file) {
        int channels, w, h;

        unsigned char* imageData = stbi_load(file.c_str(), &w, &h, &channels, 0);

        if (imageData == nullptr)
            return;

        width = w;
        height = h;
        if (channels == 3)
            format = TextureFormat::RGB;

        if (channels == 4)
            format = TextureFormat::RGBA;

        pixelData.assign(imageData, imageData + width * height * channels);
        stbi_image_free(imageData);
    }

    void SoftwareTexture::invertY() {
        unsigned int rowSize;
        if (format == TextureFormat::RGB)
            rowSize = width * 3;
        if (format == TextureFormat::RGBA)
            rowSize = width * 4;

        std::vector<unsigned char> temp(rowSize);
        for (unsigned int y = 0; y < height / 2; ++y) {
            unsigned char *topRow = &pixelData[y * rowSize];
            unsigned char *bottomRow = &pixelData[(height - y - 1) * rowSize];
            std::copy(bottomRow, bottomRow + rowSize, temp.begin());
            std::copy(topRow, topRow + rowSize, bottomRow);
            std::copy(temp.begin(), temp.end(), topRow);
        }
    }

    unsigned int SoftwareTexture::getWidth() const {
        return width;
    }

    unsigned int SoftwareTexture::getHeight() const {
        return height;
    }

    TextureFormat SoftwareTexture::getTextureFormat() {
        return format;
    }

    TextureFlag SoftwareTexture::getFlags() {
        return flags;
    }

    SoftwareTexture SoftwareTexture::downscale(unsigned int rhs) {
        std::vector<unsigned char> result;

        if (format == TextureFormat::RGB)
            result.resize((width / rhs) * (height / rhs) * 3),
            stbir_resize_uint8_linear(pixelData.data(), width, height, 0, result.data(), (width / rhs), (height / rhs), 0, stbir_pixel_layout::STBIR_RGB);

        else if (format == TextureFormat::RGBA)
            result.resize((width / rhs) * (height / rhs) * 4),
            stbir_resize_uint8_linear(pixelData.data(), width, height, 0, result.data(), (width / rhs), (height / rhs), 0, stbir_pixel_layout::STBIR_RGBA);

        return SoftwareTexture(result, format, (width / rhs), (height / rhs));
    }
}
